"""Functions to load the data and parse the annotation file."""

import re
from dataclasses import dataclass
from typing import Dict, List

import numpy as np
import pandas as pd
from tqdm import tqdm

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omerrors as errors
import omniplate.omgenutils as gu
from omniplate.parseplate import parseplate


@clogger.log
def load(
    self,
    dnames,
    anames=None,
    experimenttype=None,
    platereadertype="Tecan",
    dsheets=False,
    asheets=False,
    info=True,
    combine=False,
    combinename=None,
    min_dt=0.05,
):
    """
    Load raw data files.

    Two files are required: one generated by the plate reader and the other
    the corresponding annotation files - both assumed to be xlsx.

    Parameters
    ----------
    dnames: string or list of strings, optional
        The name of the file containing the data from the plate reader
        or a list of file names.
    anames: string or list of strings, optional
        The name of file containing the corresponding annotation or a list
        of file names.
    experimenttype: string or list of strings, optional
        If specified, creates a new experiment_type column in each
        dataframe.
    platereadertype: string or list of str
        The type of plate reader, currently either 'Tecan' or 'Sunrise'
        or 'tidy'. If you have data from different plate readers, say three,
        use a list, such as ["Tecan", "Tecan", "Sunrise"].
    dsheets: list of integers or strings, optional
        The relevant sheets of the Excel files storing the data.
    asheets: list of integers or strings, optional
        The relevant sheets of the corresponding Excel files for the
        annotation.
    info: boolean
        If True (default), display summary information on the data once
        loaded.
    combine: boolean
        If True (default is False), combine all the data loaded into one
        single experiment.
    combinename: str (optional)
        If specified, the name given to the combined experiment.
    min_dt: float, optional
        The minimum time interval taken by the plate reader between
        measurements. Used only when combine=True.

    Examples
    -------
    >>> p.load('Data.xlsx', 'DataContents.xlsx')
    >>> p.load('Data.xlsx', 'DataContents.xlsx', info= False)
    >>> p.load(['HxtA.xls', 'HxtB.xlsx'], ['HxtAcontents.xlsx',
            'HxtBcontents.xlsx'], combine=True)
    >>> p.load(['HxtG.xls', 'HxtP.xlsx'],
            ['HxtGcontents.xlsx', 'HxtPcontents.xlsx'],
            ["glucose_pregrowth", "pyruvate_pregrowth"])
    """
    (
        dnames,
        anames,
        platereadertype,
        experimenttype_dict,
        dsheets,
        asheets,
    ) = run_load_checks(
        dnames=dnames,
        anames=anames,
        experimenttype=experimenttype,
        platereadertype=platereadertype,
        dsheets=dsheets,
        asheets=asheets,
    )
    alldata = {}
    for i, (dname, platereader) in enumerate(zip(dnames, platereadertype)):
        # get dataframe for raw data
        res = loaddatafiles(
            platereadertype=platereader,
            datadirpath=self.datadirpath,
            dname=dname,
            dsheet=dsheets[i],
            aname=anames[i],
            asheet=asheets[i],
        )
        self.allexperiments.append(res.experiment)
        self.allconditions[res.experiment] = res.allconditions
        self.allstrains[res.experiment] = res.allstrains
        self.datatypes[res.experiment] = res.datatypes
        self.allstrainsconditions[res.experiment] = list(
            (res.rdf.strain + " in " + res.rdf.condition).dropna().unique()
        )
        alldata.update(res.alldata)
        # make r dataframe of raw data
        self.r = (
            pd.merge(self.r, res.rdf, how="outer")
            if hasattr(self, "r")
            else res.rdf
        )
        # update progress dictionary
        admin.initialiseprogress(self, res.experiment)
    sort_attributes(self)
    # make sc dataframe for summary stats and corrections
    alldfs = []
    for exp in alldata:
        strs, cons = [], []
        for cs in alldata[exp]:
            strs.append(cs.split(" in ")[0])
            cons.append(cs.split(" in ")[1])
        corrdict = {"experiment": exp, "strain": strs, "condition": cons}
        corrdict.update(
            {dtype + "_measured": True for dtype in self.datatypes[exp]}
        )
        alldfs.append(pd.DataFrame(corrdict))
    self.sc = pd.concat(alldfs)
    self.sc = self.sc.reset_index(drop=True)
    # replace NaN with False for datatypes_measured
    for column in self.sc.columns[self.sc.columns.str.contains("_measured")]:
        self.sc[column] = (
            self.sc[column].fillna(False).infer_objects(copy=False)
        )
    # make backup dataframe of original data
    self.origr = self.r.copy()
    # make dataframe for well contents
    self.wellsdf = admin.makewellsdf(self.r)
    # make s dataframe for summary data
    self.s = admin.make_s(self)
    # add experiment_type
    if experimenttype is not None:
        for df in [self.r, self.s, self.sc]:
            df["experiment_type"] = df.experiment.map(experimenttype_dict)
    # display info on experiment, conditions and strains
    if info:
        self.info
    print('\nWarning: wells with no strains have been changed to "Null".')
    # combine into one experiment
    if combine:
        if combinename is not None and isinstance(combinename, str):
            self.combined = combinename
        combine_experiments(self, min_dt=min_dt)


def run_load_checks(
    dnames,
    anames,
    experimenttype,
    platereadertype,
    dsheets,
    asheets,
):
    """Fix inputs to loaddata."""
    dnames = gu.makelist(dnames)
    if anames is None:
        anames = [dname.split(".")[0] + "Contents.xlsx" for dname in dnames]
    else:
        anames = gu.makelist(anames)
    if experimenttype is not None:
        experimenttype = gu.makelist(experimenttype)
        experimenttype_dict = {
            dname.split(".")[0]: e_type
            for dname, e_type in zip(dnames, experimenttype)
        }
    else:
        experimenttype_dict = None
    if isinstance(platereadertype, str):
        platereadertype = [platereadertype for dname in dnames]
    if not dsheets:
        dsheets = [0 for dname in dnames]
    if not asheets:
        asheets = [0 for dname in dnames]
    return (
        dnames,
        anames,
        platereadertype,
        experimenttype_dict,
        dsheets,
        asheets,
    )


def loaddatafiles(platereadertype, datadirpath, dname, dsheet, aname, asheet):
    """Call functions to parse data and metadate from the input files."""

    @dataclass
    class res:
        rdf: pd.DataFrame
        allconditions: List[str]
        allstrains: List[str]
        alldata: Dict
        experiment: str
        datatypes: List[str]

    experiment = dname.split(".")[0]
    # import and parse plate contents file
    allconditions, allstrains, alldata, wellcontents = analyseContentsofWells(
        datadirpath, experiment, aname, asheet
    )
    # import and parse data file created by plate reader
    try:
        print("Loading", dname)
        rdf = parseplate(dname, platereadertype, datadirpath, dsheet)
    except FileNotFoundError as exc:
        raise errors.FileNotFound(str(datadirpath / dname)) from exc
    # get datatypes
    cols = rdf.columns.to_list()
    datatypes = [col for col in cols if col != "time" and col != "well"]
    # add condition and strain to dataframe
    rdf["experiment"] = experiment
    rdf["condition"] = rdf["well"].map(
        {well: wellcontents[well][0] for well in wellcontents}
    )
    rdf["strain"] = rdf["well"].map(
        {well: wellcontents[well][1] for well in wellcontents}
    )
    # drop wells with condition=strain=None
    rdf = rdf.drop(
        rdf[rdf.condition.isnull() & rdf.strain.isnull()].index
    ).reset_index(drop=True)
    return res(
        rdf=rdf,
        allconditions=allconditions,
        allstrains=allstrains,
        alldata=alldata,
        experiment=experiment,
        datatypes=datatypes,
    )


def analyseContentsofWells(datadirpath, experiment, aname, asheetnumber):
    """
    Load and parse ContentsofWells file.

    Return wellcontents, a dictionary with the contents of each well indexed
    by well, and alldata, a dictionary describing the contents of each well
    indexed by experiment.
    """
    try:
        alldata = {}
        # import contents of the wells
        anno = pd.read_excel(
            str(datadirpath / aname), index_col=0, sheet_name=asheetnumber
        )
        alldata[experiment] = []
        wellcontents = {}
        # run through and parse content of each well
        for x in np.arange(1, 13):
            for y in "ABCDEFGH":
                well = y + str(x)
                if (
                    isinstance(anno[x][y], str)
                    and anno[x][y] != "contaminated"
                ):
                    s, c = anno[x][y].split(" in ")
                    # standardise naming of wells with no strains
                    s = re.sub("(null|NULL)", "Null", s)
                    wellcontents[well] = [c.strip(), s.strip()]
                    alldata[experiment].append(
                        wellcontents[well][1] + " in " + wellcontents[well][0]
                    )
                else:
                    wellcontents[well] = [None, None]
        # create summary descriptions of the well contents
        alldata[experiment] = list(np.unique(alldata[experiment]))
        allconditions = list(
            np.unique(
                [
                    wellcontents[well][0]
                    for well in wellcontents
                    if wellcontents[well][0] is not None
                ]
            )
        )
        allstrains = list(
            np.unique(
                [
                    wellcontents[well][1]
                    for well in wellcontents
                    if wellcontents[well][0] is not None
                ]
            )
        )
        return allconditions, allstrains, alldata, wellcontents
    except FileNotFoundError as exc:
        raise errors.FileNotFound(str(datadirpath / aname)) from exc


def combine_experiments(self, min_dt, dt=None):
    """
    Combine raw data from all loaded experiments into one experiment.

    Only to be run immediately after first loading the data.

    Wells are relabelled with a prefix denoting their original
    experiment, and p.experiment_map gives the mapping back to the
    original experiment names.
    """
    print("Combining experiments...")
    # to make new r dataframe
    rdict = {col: [] for col in list(self.r.columns)}
    # find data types present in all experiments
    datatypes = self.r.select_dtypes(include="number").columns.tolist()
    dfindices = self.r.select_dtypes(include=["object"]).columns.tolist()
    # define a new universal time
    datatypes.remove("time")
    unique_t = np.sort(self.r.time.unique())
    if dt is None:
        ut_values, ut_counts = np.unique(
            np.round(np.diff(unique_t), 2), return_counts=True
        )
        dt = ut_values[ut_values > min_dt][
            np.argmax(ut_counts[ut_values > min_dt])
        ]
    print(f" Using dt = {dt:.2f}.")
    nt = np.arange(0, unique_t.max(), dt)
    # find index for all wells
    wells_index = list(self.r.groupby(dfindices).mean().index)
    # interpolate data for each well to new universal time
    print(" Interpolating data to a universal time.")
    for wi in tqdm(wells_index):
        df = self.r.query(
            " and ".join(
                [
                    f"{column} == '{value}'"
                    for column, value in zip(dfindices, wi)
                ]
            )
        )
        # local universal time depending on length of experiments
        l_nt = nt[
            np.argmin((df.time.min() - nt) ** 2) : np.argmin(
                (df.time.max() - nt) ** 2
            )
            + 1
        ]
        x = df.time.values
        xs = np.sort(x)
        # interpolate data for this well to universal time
        new_y = {}
        for datatype in datatypes:
            try:
                y = df[datatype].values
                ys = y[np.argsort(x)]
                new_y[datatype] = np.interp(l_nt, xs, ys)
            except KeyError:
                new_y[datatype] = np.nan * np.ones(l_nt.shape)
        # add to dict
        for i in range(l_nt.size):
            for column, value in zip(dfindices, wi):
                rdict[column].append(value)
            rdict["time"].append(l_nt[i])
            for datatype in datatypes:
                rdict[datatype].append(new_y[datatype][i])
    # re-make r dataframe, now with universal time
    self.r = pd.DataFrame(data=rdict)
    # make dict mapping experiment number to original name
    experiment_map = {
        expt: str(i) for i, expt in enumerate(self.r.experiment.unique())
    }
    # rename wells
    self.r.well = (
        self.r.experiment.astype("string").replace(experiment_map)
        + "_"
        + self.r.well
    )
    self.r["well"] = self.r["well"].astype("object")
    # save names of original experiments
    self.r["original_experiment"] = self.r.experiment.copy()
    self.r["experiment_id"] = (
        self.r.experiment.astype("string").replace(experiment_map).copy()
    )
    self.r["experiment_id"] = self.r["experiment_id"].astype("object")
    # write over r.experiment and sort
    self.r.experiment = self.combined
    self.r = self.r.sort_values(by=["time", "well"]).reset_index(drop=True)
    # remake summary attributes
    self.allexperiments = [self.combined]
    for mattr in [
        "allconditions",
        "allstrains",
        "allstrainsconditions",
        "datatypes",
    ]:
        uniques = set(
            [
                lattr
                for e in getattr(self, mattr)
                for lattr in getattr(self, mattr)[e]
            ]
        )
        setattr(self, mattr, {self.combined: list(uniques)})
    sort_attributes(self)
    # reset self.progress
    for key in self.progress:
        self.progress[key] = {
            self.combined: self.progress[key][next(iter(self.progress[key]))]
        }
    # remake dataframe for well content
    self.wellsdf = admin.makewellsdf(self.r)
    # remake s dataframe
    self.s = admin.make_s(self)
    # add original experiments to sc dataframe
    self.sc["original_experiment"] = self.sc.experiment.copy()
    self.sc["experiment_id"] = (
        self.sc.experiment.astype("string").replace(experiment_map).copy()
    )
    self.sc.experiment = self.combined


def sort_attributes(self):
    """Alphabetically sort attributes describing experiments."""
    self.allexperiments = sorted(self.allexperiments)
    for e in self.allconditions:
        self.allconditions[e] = sorted(
            self.allconditions[e], key=lambda x: (x.split()[-1], x.split()[0])
        )
    for e in self.allstrains:
        self.allstrains[e] = sorted(self.allstrains[e])
    for e in self.allstrainsconditions:
        self.allstrainsconditions[e] = sorted(
            self.allstrainsconditions[e],
            key=lambda x: (
                x.split(" in ")[0],
                x.split(" in ")[1].split()[-1],
                x.split(" in ")[1].split()[0],
            ),
        )



import re

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omerrors as errors
import omniplate.sunder as sunder


@clogger.log
def rename_combined(self, newname):
    """Rename a combined experiment."""
    if self.allexperiments == [self.combined]:
        for df in [self.r, self.s, self.sc]:
            df["experiment"] = df.experiment.replace({self.combined: newname})
        self.allexperiments = [newname]
        self.combined = newname


def search(self, listofstrings):
    """
    Search self.allstrainsconditions.

    Example
    -------
    >>> p.search(["HXT6", "Glu"])
    """
    if isinstance(listofstrings, str):
        listofstrings = [listofstrings]
    for e in self.allstrainsconditions:
        for sc in self.allstrainsconditions[e]:
            hits = [q for q in listofstrings if q in sc]
            if len(hits) == len(listofstrings):
                print(f"{e}: {sc}")


@clogger.log
def rename(self, translatedict, regex=True, **kwargs):
    """
    Rename strains or conditions.

    Uses a dictionary to replace all occurrences of a strain or a condition
    with an alternative.

    Note that instances of self.progress will not be updated.

    Parameters
    ----------
    translatedict: dictionary
        A dictionary of old name - new name pairs
    regex: boolean (optional)
        Value of regex to pass to panda's replace.
    kwargs: keyword arguments
        Passed to panda's replace.

    Example
    -------
    >>> p.rename({'77.WT' : 'WT', '409.Hxt4' : 'Hxt4'})
    """
    # check for duplicates
    if (
        len(translatedict.values())
        != np.unique(list(translatedict.values())).size
    ):
        print("Warning: new names are not unique.")
        # replace in dataframes
        for df in [self.r, self.sc]:
            exps = df.experiment.copy()
            df.replace(translatedict, inplace=True, regex=regex, **kwargs)
            # do not change names of experiments
            df["experiment"] = exps
        # remake s so that strains with the same name are combined
        self.s = admin.make_s(self)
        # remake sc
        self.sc = self.sc.drop_duplicates().reset_index(drop=True)
    else:
        # replace in dataframes
        for df in [self.r, self.s, self.sc]:
            exps = df.experiment.copy()
            df.replace(translatedict, inplace=True, regex=regex, **kwargs)
            df["experiment"] = exps
    self.wellsdf = admin.makewellsdf(self.r)
    # replace in attributes - allstrains and allconditions
    for e in self.allexperiments:
        for listattr in [self.allconditions[e], self.allstrains[e]]:
            for i, listitem in enumerate(listattr):
                for key in translatedict:
                    if key in listitem:
                        listattr[i] = listitem.replace(key, translatedict[key])
        # unique values in case two strains have been renamed to one
        self.allconditions[e] = sorted(list(np.unique(self.allconditions[e])))
        self.allstrains[e] = sorted(list(np.unique(self.allstrains[e])))
        self.allstrainsconditions[e] = list(
            (self.r.strain + " in " + self.r.condition).dropna().unique()
        )


@clogger.log
def addcolumn(self, newcolumnname, oldcolumn, newcolumnvalues):
    """
    Add a new column to all dataframes by parsing an existing column.

    All possible entries for the new column are specified as strings and
    the entry in the new column will be whichever of these strings is
    present in the entry of the existing column.

    Parameters
    ----------
    newcolumnname: string
        The name of the new column.
    oldcolumn: string
        The name of the column to be parsed to create the new column.
    newcolumnvalues: list of strings
        All of the possible values for the entries in the new column.

    Example
    -------
    >>> p.addcolumn('medium', 'condition', ['Raffinose',
    ...                                     'Geneticin'])

    will parse each entry in 'condition' to create a new column called
    'medium' that has either a value 'Raffinose' if 'Raffinose' is in the
    entry from 'condition' or a value 'Geneticin' if 'Geneticin' is in the
    entry from 'condition'.
    """
    for dftype in ["r", "s", "sc"]:
        df = getattr(self, dftype)
        oldcol = df[oldcolumn].to_numpy()
        newcol = np.array(("none",) * oldcol.size, dtype="object")
        for i, oldcolvalue in enumerate(oldcol):
            for newcolvalue in newcolumnvalues:
                if newcolvalue in oldcolvalue:
                    newcol[i] = newcolvalue
        newcolseries = pd.Series(newcol, index=df.index, name=newcolumnname)
        setattr(self, dftype, pd.concat([df, newcolseries], axis=1))


@clogger.log
def addnumericcolumn(
    self,
    newcolumnname,
    oldcolumn,
    picknumber=0,
    leftsplitstr=None,
    rightsplitstr=None,
    asstr=False,
):
    """
    Add a new numeric column.

    Parse the numbers from the entries of an existing column.

    Run only after the basic analyses - ignorewells, correctOD, and
    correctmedia - have been performed because addnumericolumn changes
    the structure of the dataframes.

    Parameters
    ----------
    newcolumnname: string
        The name of the new column.
    oldcolumn: string
        The name of column to be parsed.
    picknumber: integer
        The number to pick from the list of numbers extracted from the
        existing column's entry.
    leftsplitstr: string, optional
        Split the entry of the column using whitespace and parse numbers
        from the substring to the immediate left of leftsplitstr rather
        than the whole entry.
    rightsplitstr: string, optional
        Split the entry of the column using whitespace and parse numbers
        from the substring to the immediate right of rightsplitstr rather
        than the whole entry.
    asstr: boolean
        If True, convert the numeric value to a string to improve plots
        with seaborn.

    Examples
    --------
    To extract concentrations from conditions use

    >>> p.addnumericcolumn('concentration', 'condition')

    For a condition like '0.5% Raf 0.05ug/mL Cycloheximide', use

    >>> p.addnumericcolumn('raffinose', 'condition',
    ...                     picknumber= 0)
    >>> p.addnumericcolumn('cycloheximide', 'condition',
    ...                     picknumber= 1)
    """
    # process splitstrs
    if leftsplitstr or rightsplitstr:
        splitstr = leftsplitstr if leftsplitstr else rightsplitstr
        locno = -1 if leftsplitstr else 1
    else:
        splitstr = False
    # change each dataframe
    for dftype in ["r", "s", "sc"]:
        df = getattr(self, dftype)
        if asstr:
            # new column of strings
            newcol = np.full_like(df[oldcolumn].to_numpy(), "", dtype="object")
        else:
            # new column of floats
            newcol = np.full_like(
                df[oldcolumn].to_numpy(), np.nan, dtype="float"
            )
        # parse old column
        for i, oldcolvalue in enumerate(df[oldcolumn].to_numpy()):
            if oldcolvalue:
                # split string first on spaces and then find substring
                # adjacent to specified splitstring
                if splitstr:
                    if splitstr in oldcolvalue:
                        # oldcolvalue contains leftsplitstring or
                        # rightsplitstring
                        bits = oldcolvalue.split()
                        for k, bit in enumerate(bits):
                            if splitstr in bit:
                                loc = k + locno
                                break
                        # adjacent string
                        oldcolvalue = bits[loc]
                    else:
                        # oldcolvalue does not contain leftsplitstring
                        # or rightsplitstring
                        oldcolvalue = ""
                # loop through all floats in oldcolvalue
                nocount = 0
                for ci in re.split(
                    r"[+-]?([0-9]+(?:[.][0-9]*)?|[.][0-9]+)", oldcolvalue
                ):
                    try:
                        no = float(ci)
                        if nocount == picknumber:
                            newcol[i] = ci if asstr else no
                            break
                        nocount += 1
                    except ValueError:
                        pass
        newcolseries = pd.Series(newcol, index=df.index, name=newcolumnname)
        setattr(self, dftype, pd.concat([df, newcolseries], axis=1))


@clogger.log
def add_to_sc(
    self,
    newcolumn=None,
    s_column=None,
    func=None,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
    strains="all",
    strainincludes=False,
    strainexcludes=False,
):
    """
    Apply func to a column in the s dataframe.

    The results are stored in the sc dataframe.

    Parameters
    ----------
    newcolumn:  string
        The name of the new column in the sc dataframe
    s_column:   string
        The name of the column in s dataframe from which the
        data is to be processed
    func:   function
        The function to be applied to the data in the s dataframe.

    Examples
    --------
    >>> p.add_to_sc(newcolumn= "max_GFP", s_column= "mean_GFP",
    ...             func= np.nanmax)
    >>> p.add_to_sc(newcolumn= "lower_quartile_GFP", s_column= "mean_GFP",
    ...             func= lambda x: np.nanquantile(x, 0.25))
    """
    # extract data
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull=True,
    )
    self.sc[newcolumn] = np.nan
    for e in exps:
        for c in cons:
            for s in strs:
                d = self.s.query(
                    "experiment == @e and condition == @c and strain == @s"
                )[s_column].values
                res = np.asarray(func(d))
                if res.size == 1:
                    self.sc.loc[
                        (self.sc.experiment == e)
                        & (self.sc.condition == c)
                        & (self.sc.strain == s),
                        newcolumn,
                    ] = func(d)
                else:
                    print("func must return a single value.")
                    return False


@clogger.log
def addcommonvar(
    self,
    var="time",
    dvar=None,
    figs=True,
    experiments="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditions="all",
    conditionincludes=False,
    conditionexcludes=False,
    strains="all",
    strainincludes=False,
    strainexcludes=False,
):
    """
    Add a common variable to all time-dependent dataframes.

    The common variable allows averaging across experiments
    and typically is time.

    A common variable is added to time-dependent dataframes. This
    variable's values only come from a fixed array so that they are
    from the same array for all experiments.

    For example, the plate reader often does not perfectly increment time
    between measurements and different experiments can have slightly
    different time points despite the plate reader having the same
    settings. These unique times prevent seaborn from taking averages.

    If experiments have measurements that start at the same time point and
    have the same interval between measurements, then setting a commontime
    for all experiments will allow seaborn to perform averaging.

    The array of the common variable has an interval dvar, which is
    automatically calculated, but may be specified.

    Each instance of var is assigned a common value - the closest instance
    of the common variable to the instance of var. Measurements are assumed
    to the same for the true instance of var and for the assigned common
    value, which may generate errors if these two are sufficiently
    distinct.

    An alternative method is averageoverexpts.

    Parameters
    ----------
    var: string
        The variable from which the common variable is generated,
        typically 'time'.
    dvar: float, optional
        The interval between the values comprising the common array.
    figs: boolean
        If True, generate plot to check if the variable and the common
        variable generated from it are sufficiently close in value.
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in their
        name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.

    Example
    -------
    To plot averages of time-dependent variables over experiments, use for
    example

    >>> p.addcommonvar('time')
    >>> p.plot(x= 'commontime', y= 'cGFPperOD', hue= 'condition')
    """
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull=False,
    )
    for i, df in enumerate([self.r, self.s]):
        if var in df:
            loc = (
                df.experiment.isin(exps)
                & df.condition.isin(cons)
                & df.strain.isin(strs)
            )
            if dvar is None and df.equals(self.r):
                # group data by experiment generating average time
                mdf = (
                    df[loc]
                    .sort_values(by=["experiment", "well", var])[
                        ["experiment", "well", var, "OD"]
                    ]
                    .groupby(["experiment", var])
                    .mean(numeric_only=True)
                ).reset_index()[["experiment", var]]
                # find experiment with longest duration
                longest_i = np.argmax(
                    [mdf[mdf.experiment == e][var].size for e in exps]
                )
                # chose the var for this longest experiment
                cvar = mdf[mdf.experiment == exps[longest_i]][var].values
            elif dvar is not None:
                cvar = np.arange(df[loc][var].min(), df[loc][var].max(), dvar)
            # define common var
            df.loc[loc, f"common{var}"] = df[loc][var].apply(
                lambda x: cvar[np.argmin((x - cvar) ** 2)]
            )
            if figs and i == 0:
                plt.figure()
                sl = np.linspace(
                    df[loc][var].min(), 1.05 * df[loc][var].max(), 100
                )
                plt.plot(sl, sl, alpha=0.4)
                plt.plot(
                    df[loc][var].to_numpy(),
                    df[loc]["common" + var].to_numpy(),
                    ".",
                )
                plt.xlabel(var)
                plt.ylabel("common" + var)
                title = "r dataframe" if df.equals(self.r) else "s dataframe"
                plt.title(title)
                plt.suptitle(
                    f"comparing {var} with common{var} – "
                    "the line y= x is expected."
                )
                plt.tight_layout()
                plt.show(block=False)


@clogger.log
def restricttime(self, tmin=None, tmax=None):
    """
    Restrict the processed data to a range of time.

    Points outside this time range are ignored.

    Note that data in the dataframes outside the time range is lost.
    Exporting the dataframes before running restricttime is recommended.

    Parameters
    ----------
    tmin: float
        The minimum value of time, with data kept only for t >= tmin.
    tmax: float
        The maximum value of time, with data kept only for t <= tmax.

    Example
    -------
    >>> p.restricttime(tmin= 5)
    """
    if tmin is None:
        tmin = self.r.time.min()
    if tmax is None:
        tmax = self.r.time.max()
    if tmax > tmin:
        self.r = self.r[(self.r.time >= tmin) & (self.r.time <= tmax)]
        self.s = self.s[(self.s.time >= tmin) & (self.s.time <= tmax)]
    else:
        print("tmax or tmin is not properly defined.")


@clogger.log
def getdataframe(
    self,
    dfname="s",
    experiments="all",
    conditions="all",
    strains="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditionincludes=False,
    conditionexcludes=False,
    strainincludes=False,
    strainexcludes=False,
    nonull=True,
):
    """
    Obtain a subset of the data in a dataframe.

    This data can be used plotted directly.

    Parameters
    ---------
    dfname: string
        The dataframe of interest either 'r' (raw data),
        's' (default; processed data),
        or 'sc' (summary statistics).
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in their
        name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.
    nonull: boolean, optional
        If True, ignore 'Null' strains

    Returns
    -------
    ndf: dataframe

    Example
    -------
    >>> ndf= p.getdataframe('s', conditions= ['2% Glu'],
    ...                     nonull= True)
    """
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull,
    )
    if hasattr(self, dfname):
        df = getattr(self, dfname)
        ndf = df.query(
            "experiment == @exps and condition == @cons " "and strain == @strs"
        )
        if ndf.empty:
            print("No data found.")
        else:
            return ndf.copy()
    else:
        raise errors.UnknownDataFrame(
            "Dataframe " + dfname + " is not recognised."
        )



"""
A Bayesian method to correct for autofluorescence.

The GFP channel is denoted y ; the reference strain is denoted WT.
"""

import concurrent.futures
import os
from dataclasses import asdict, dataclass, fields

import matplotlib.pylab as plt
import numpy as np
from tqdm import tqdm
from scipy.optimize import minimize_scalar

import omniplate.corrections as omcorr
import omniplate.omplot as omplot
import omniplate.sunder as sunder
from omniplate.omfitderiv import runfitderiv

rng = np.random.default_rng()
min_fl = 1e-4


def bootstrap_array(a):
    """Bootstrap 1D array."""
    na = a[rng.integers(0, a.size, a.size)]
    return na


@dataclass
class local_data:
    """Data from one time point."""

    ly: float
    lywt: float
    lyn: float
    od: float
    odwt: float

    def to_array(self):
        """Convert to array."""
        return np.array(list(asdict(self).values()))

    def to_list(self):
        """Convert to list."""
        return list(asdict(self).values())

    @classmethod
    def boot_from_list(cls, ell):
        """Define and bootstrap sample from an array."""
        field_names = [f.name for f in fields(cls)]
        return cls(
            **{
                name: bootstrap_array(val)
                for name, val in zip(field_names, ell)
            }
        )


@dataclass
class theta:
    """Parameters to infer."""

    a: float
    g: float

    def to_array(self):
        """Convert to array."""
        return np.array(list(asdict(self).values()))

    @classmethod
    def from_array(cls, arr):
        """Define from an array."""
        field_names = [f.name for f in fields(cls)]
        return cls(**{name: val for name, val in zip(field_names, arr)})

    @classmethod
    def zeros(cls):
        """Define with all values as zeros."""
        field_names = [f.name for f in fields(cls)]
        return cls(
            **{
                name: val
                for name, val in zip(field_names, np.zeros(len(field_names)))
            }
        )


@dataclass
class suff_stats:
    """Sufficient statistics."""

    yT: float
    ywtT: float
    lyT: float
    lywtT: float


def estimate_by(ld):
    """Find ML estimate of by."""
    return np.exp(np.mean(ld.lyn))


def find_suff_stats(x, ld):
    """Define the sufficient statistics."""
    th = theta.from_array(x)
    by = estimate_by(ld)
    yT = ld.od * (th.a + th.g) + by
    ywtT = ld.odwt * th.a + by
    lyT = np.log(yT)
    lywtT = np.log(ywtT)
    res = suff_stats(
        yT=yT,
        lyT=lyT,
        ywtT=ywtT,
        lywtT=lywtT,
    )
    return res


def find_bounds_x0(ld):
    """Find bounds on theta and make an initial estimate."""
    y = np.exp(ld.ly)
    ywt = np.exp(ld.lywt)
    by = estimate_by(ld)
    odwt = np.mean(ld.odwt)
    od = np.mean(ld.od)
    # bounds
    bounds = {"a": (min_fl, np.max(ywt) / odwt), "g": (min_fl, np.max(y) / od)}
    bounds = [
        bounds[theta_name] for theta_name in [f.name for f in fields(theta)]
    ]
    # estimate theta: a and g
    a0 = np.max([min_fl, (np.exp(np.mean(ld.lywt)) - by) / odwt])
    g0 = np.max([min_fl, (np.exp(np.mean(ld.ly)) - by) / od - a0])
    x0 = [a0, g0]
    return bounds, x0


def a_error(a, ld, params):
    """Find term in -log(posterior) from wild-type data."""
    ss = find_suff_stats([a, np.nan], ld)
    err = np.sum((ld.lywt - ss.lywtT) ** 2)
    return err


def g_error(g, ld, a_opt, params):
    """Find term in -log(posterior) from tagged-strain data."""
    ss = find_suff_stats(np.array([a_opt, g]), ld)
    err = np.sum((ld.ly - ss.lyT) ** 2)
    return err


def find_mode(ld, params):
    """Find mode of posterior using two 1d minimisations."""
    bounds, _ = find_bounds_x0(ld)
    a_opt = np.nan
    g_opt = np.nan
    ares = minimize_scalar(
        a_error, bounds=bounds[0], method="bounded", args=(ld, params)
    )
    if ares.success:
        a_opt = ares.x
        gres = minimize_scalar(
            g_error,
            bounds=bounds[1],
            method="bounded",
            args=(ld, a_opt, params),
        )
        if gres.success:
            g_opt = gres.x
    return theta(a=a_opt, g=g_opt)


def process_strain(strain, params):
    """Correct autofluorescence for one strain."""
    t = params["t"]
    # get data for tagged strain
    if params["y_data_dict"][strain] is None:
        return
    else:
        try:
            y, od = params["y_data_dict"][strain]
        except Exception as ex:
            print(f"Error accessing data for strain {strain}: {ex}")
            return
    print(
        f"{params['experiment']}: {strain} in {params['condition']};"
        f" {y.shape[1]} replicates"
    )
    flperod_raw = np.full((t.size, params["noboots"]), np.nan)
    # correct autofluorescence for each time point
    for i_boot in range(params["noboots"]):
        for i in range(t.size):
            ld = local_data(
                ly=np.log(y[i, :]),
                lywt=np.log(params["ywt"][i, :]),
                lyn=np.log(params["yn"][i, :]),
                od=od[i, :],
                odwt=params["odwt"][i, :],
            )
            if i_boot > 0:
                ld = local_data.boot_from_list(ld.to_list())
            mode = find_mode(ld, params)
            flperod_raw[i, i_boot] = mode.g
    return strain, flperod_raw


def process_strains_parallel(params, allstrains, max_workers=None):
    """Correct multiple strains in parallel."""
    results = []
    with concurrent.futures.ProcessPoolExecutor(
        max_workers=max_workers
    ) as executor:
        # submit all tasks and collect futures as dict
        futures_dict = {
            executor.submit(process_strain, s, params): s for s in allstrains
        }
        # process results as they complete
        for future in tqdm(
            concurrent.futures.as_completed(futures_dict),
            total=len(futures_dict),
            desc="Processing strains",
        ):
            strain = futures_dict[future]
            try:
                result = future.result()
                if result is not None:
                    results.append(result)
            except Exception as exc:
                print(f"{strain} generated an exception: {exc}")
    return results


def define_bdn(options):
    """Define hyperparameters for smoothing Gaussian process."""
    bd_default = {0: (-2, 8), 1: (0, 4), 2: (-2, 1)}
    if options["bd"] is not None:
        bdn = bd_default | options["bd"]
    else:
        bdn = bd_default
    return bdn


def correctauto_bayesian(
    self,
    f,
    refstrain,
    noprocessors,
    experiments,
    experimentincludes,
    experimentexcludes,
    conditions,
    conditionincludes,
    conditionexcludes,
    strains,
    strainincludes,
    strainexcludes,
    options,
):
    """
    Correct fluorescence for auto- and background fluorescence.

    Use a Bayesian method, with

        y = exp(epsilon) * ((g + a)*od + by)
        ywt = exp(epsilon) * (a*odwt + by)

    with epsilon normally distributed and modelling measurement noise.
    """
    print(f"Using Bayesian approach for {f[0]}.")
    bdn = define_bdn(options)
    for e in sunder.getset(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        labeltype="experiment",
        nonull=True,
    ):
        for c in sunder.getset(
            self,
            conditions,
            conditionincludes,
            conditionexcludes,
            labeltype="condition",
            nonull=True,
            nomedia=True,
        ):
            # get data for reference strain
            t, (ywt, odwt) = sunder.extractwells(
                self.r, self.s, e, c, refstrain, [f[0], "OD"]
            )
            ywt, keep = de_nan(ywt)
            t = t[keep]
            odwt = odwt[keep, :]
            odwt[odwt < 0] = np.min(odwt[odwt > 0])
            # get data for Null
            _, yn = sunder.extractwells(self.r, self.s, e, c, "Null", f)
            yn, _ = de_nan(yn)
            if not sufficient_replicates(ywt, yn, refstrain, c):
                continue
            # use all yn to estimate sy2
            sy2 = np.max([1e-4, np.var(np.log(yn))])
            params = {
                "experiment": e,
                "condition": c,
                "t": t,
                "ywt": ywt,
                "yn": yn,
                "odwt": odwt,
                "sy2": sy2,
                "bdn": bdn,
                "y_data_dict": {},
            }
            params = params | options
            # get strains
            allstrains = sunder.getset(
                self,
                strains,
                strainincludes,
                strainexcludes,
                labeltype="strain",
                nonull=True,
            )
            allstrains = [
                s
                for s in allstrains
                if f"{s} in {c}" in self.allstrainsconditions[e]
                and s != refstrain
            ]
            # define data before parallel processing
            for s in allstrains:
                y, od, success = get_tagged_strain_data(self, e, c, s, f, keep)
                if success:
                    od[od < 0] = np.min(od[od > 0])
                    params["y_data_dict"][s] = (y, od)
                else:
                    params["y_data_dict"][s] = None
            print(f"{c}: Processing {len(allstrains)} strains.")
            # correct autofluorescence
            noprocessors = np.min([noprocessors, os.cpu_count() - 1])
            if noprocessors > 1:
                # parallel
                print(
                    f"Using {noprocessors} out of {os.cpu_count()} processors."
                )
                results = process_strains_parallel(
                    params, allstrains, noprocessors
                )
            else:
                # sequential
                results = [
                    process_strain(s, params)
                    for s in allstrains
                    if params["y_data_dict"][s] is not None
                ]
            # finish up
            finish_store_plot(self, f, e, c, s, t, results, params)


def finish_store_plot(
    self,
    f,
    e,
    c,
    s,
    t,
    results,
    params,
):
    """Smooth, add results to s-dataframe, and plot."""
    for s, flperod_raw in results:
        flperod_samples, f_fitderiv = smooth_with_GPs(
            t,
            flperod_raw,
            bd=params["bdn"],
            flcvfn=params["flcvfn"],
            nosamples=params["nosamples"],
            experiment=e,
            condition=c,
            strain=s,
            maxdatapts=params["maxdatapts"],
        )
        store_results(
            self,
            f,
            e=e,
            c=c,
            s=s,
            t=t,
            flperod_samples=flperod_samples,
            flperod_raw=flperod_raw,
            corrected_name="bc" + f[0],
        )
        print(f"{e}: {s} in {c}")
        f_fitderiv.gp.results()
        print("---")
        if params["figs"]:
            omplot.inspect(
                self,
                fl=f[0],
                ylabel=f"bc{f[0]}perOD",
                strains=s,
                conditions=c,
                experiments=e,
            )
        if params["fitderiv_figs"]:
            f_fitderiv.plotfit(e, c, s, "flperOD", "d/dtflperOD", logs=False)


def plot_noise(p, f, experiment, condition, strain, refstrain):
    """Plot noise in y-channel from omniplate instance."""
    t, ywt = sunder.extractwells(p.r, p.s, experiment, condition, refstrain, f)
    ywt, od_keep = de_nan(ywt)
    t = t[od_keep]
    # get data for Null
    _, yn = sunder.extractwells(p.r, p.s, experiment, condition, "Null", f)
    yn, _ = de_nan(yn)
    # get data for tagged strain
    y, od, _ = get_tagged_strain_data(
        p, experiment, condition, strain, f, od_keep, warn=False
    )
    title = f"{experiment}: {strain} in {condition}"
    plt.figure(figsize=(10, 6))
    plt.subplot(3, 1, 1)
    plt.plot(t, np.var(np.log(y), axis=1), "o", label="y")
    plt.plot(t, np.var(np.log(ywt), axis=1), ".", label="ywt")
    plt.plot(t, np.var(np.log(yn), axis=1), ".", label="yn")
    plt.grid()
    plt.yscale("log")
    plt.ylabel("Var(log fl)")
    plt.legend()
    plt.subplot(3, 1, 2)
    for j in range(y.shape[1]):
        plt.plot(t, y[:, j], color="b")
    plt.grid()
    plt.ylabel("fluorescence")
    plt.xlabel("time")
    plt.subplot(3, 1, 3)
    for j in range(od.shape[1]):
        plt.plot(t, y[:, j])
    plt.grid()
    plt.ylabel("OD")
    plt.xlabel("time")
    plt.suptitle(title)
    plt.tight_layout()
    plt.show(block=False)


def store_results(
    self, f, e, c, s, t, flperod_samples, flperod_raw, corrected_name
):
    """Create dict of results and add to data frames."""
    # store results for fluorescence per OD
    autofdict = {
        "experiment": e,
        "condition": c,
        "strain": s,
        "time": t,
    }
    autofdict[f"{corrected_name}perOD"] = np.mean(flperod_samples, 1)
    autofdict[f"{corrected_name}perOD_err"] = 2 * np.std(flperod_samples, 1)
    if flperod_raw.ndim == 1:
        flperod_raw = flperod_raw.reshape(flperod_raw.size, 1)
    autofdict[f"flperOD_raw_{f[0]}"] = flperod_raw[:, 0]
    omcorr.addtodataframes(self, autofdict)


def de_nan(y, most=True):
    """Remove NaN by discarding some replicates."""
    # NaNs are generated because experiments have different durations
    allnancols = np.all(np.isnan(y), axis=0)
    y = y[:, ~allnancols]
    counts = np.array(
        [np.where(~np.isnan(y[i, :]))[0].size for i in range(y.shape[0])]
    )
    if most:
        # choose the most replicates per time point
        keep = counts == counts.max()
    else:
        # choose the replicates with the longest duration
        keep = counts == counts[np.nonzero(counts)[0][-1]]
    return y[keep, :], keep


def sufficient_replicates(ywt, yn, refstrain, c):
    """Check sufficient numbers of replicates."""
    sufficient = True
    if ywt.shape[1] < 3:
        print(
            f"There are only {ywt.shape[1]} replicates (<3)"
            f" for the {refstrain} strain in {c}."
        )
        print("Abandoning correctauto.")
        sufficient = False
    if yn.shape[1] < 3:
        print(
            f"There are only {yn.shape[1]} replicates (<3)"
            f" for the Null strain in {c}."
        )
        print("Abandoning correctauto.")
        sufficient = False
    return sufficient


def get_tagged_strain_data(self, e, c, s, f, od_keep, warn=True):
    """Get and check data for fluorescently tagged strain."""
    success = True
    # get data for tagged strain
    if len(f) == 2:
        _, (y, z, od) = sunder.extractwells(
            self.r, self.s, e, c, s, f.copy() + ["OD"]
        )
        y, z, _ = de_nan(y, z)
    elif len(f) == 1:
        _, (y, od) = sunder.extractwells(
            self.r, self.s, e, c, s, f.copy() + ["OD"]
        )
        y, _ = de_nan(y)
    # make OD match t
    od = od[np.nonzero(od_keep)[0], :]
    if y.size == 0 or od.size == 0:
        if warn:
            print(f"Warning: No data found for {e}: {s} in {c}!!")
        success = False
    if y.shape[1] < 3:
        if warn:
            print(
                f"There are only {y.shape[1]} replicates (<3)"
                f" for {e}: {s} in {c}."
            )
            print("Abandoning correctauto.")
        success = False
    if len(f) == 2:
        return y, z, od, success
    elif len(f) == 1:
        return y, od, success


def smooth_with_GPs(
    t,
    y,
    bd,
    nosamples,
    experiment,
    condition,
    strain,
    flcvfn,
    maxdatapts,
):
    """Smooth y, typically raw flperod, using a GP."""
    if np.any(y[~np.isnan(y)]):
        # run GP: omfitderiv deals with NaNs
        fitvar = "log(flperOD)"
        f_fitderiv = runfitderiv(
            t,
            y,
            fitvar,
            f"d/dt_{fitvar}",
            experiment=experiment,
            condition=condition,
            strain=strain,
            bd=bd,
            cvfn=flcvfn,
            logs=False,
            negs2nan=False,
            maxdatapts=maxdatapts,
            plotlocalmax=False,
        )
        if not f_fitderiv.success:
            print(
                "-> Smoothing failed for "
                f"{experiment}: {strain} in {condition}."
            )
            return np.nan * np.ones((t.size, nosamples))
        y_samples = f_fitderiv.fitderivsample(nosamples)[0]
    else:
        print("No positive data.")
        # return all NaN
        y_samples = np.nan * np.ones((t.size, nosamples))
    return y_samples, f_fitderiv



"""Miscellaneous general functions."""

import numpy as np
import pandas as pd


def merge_df_into(df, newdf, key_cols):
    """Add newdf to df using key_cols and updating where necessary."""
    # merge preserving order of rows in df
    df["temp_index"] = range(len(df))
    merged_df = pd.merge(
        df,
        newdf,
        how="outer",
        on=key_cols,
        suffixes=("", "_update"),
    )
    merged_df = (
        merged_df.sort_values("temp_index")
        .drop("temp_index", axis=1)
        .reset_index(drop=True)
    )
    df.drop("temp_index", axis=1, inplace=True)
    # update columns in df
    cols_to_update = [col for col in newdf.columns if col not in key_cols]
    for col in cols_to_update:
        if col in df.columns:
            df.update({col: merged_df[col + "_update"]})
    # add any new columns to df
    new_cols = [
        col
        for col in newdf.columns
        if col not in df.columns and col not in key_cols
    ]
    for col in new_cols:
        df[col] = merged_df[col]


@property
def cols_to_underscore(self):
    """Replace spaces in column names of all dataframes with underscores."""
    for df in [self.r, self.s, self.sc]:
        df.columns = df.columns.str.replace(" ", "_")


def rm_under(str):
    """Replace underscore in a string with a space."""
    return str.replace("_", " ")


def mergedicts(original, update):
    """
    Merge two dicts into one.

    Parameters
    --
    original: first dict
    update: second dict
    """
    if update is not None:
        z = original | update
        return z
    else:
        return original


def findsmoothvariance(y, filtsig=0.1, nopts=False):
    """
    Estimate and then smooth the variance over replicates of data.

    Parameters
    --
    y: data - one column for each replicate
    filtsig: sets the size of the Gaussian filter used to smooth the variance
    nopts: if set, uses estimateerrorbar to estimate the variance
    """
    from scipy.ndimage import filters

    if y.ndim == 1:
        # one dimensional data
        v = estimateerrorbar(y, nopts) ** 2
    else:
        # multi-dimensional data
        v = np.var(y, 1)
    # apply Gaussian filter
    vs = filters.gaussian_filter1d(v, int(len(y) * filtsig))
    return vs


def estimateerrorbar(y, nopts=False):
    """
    Estimate measurement error for each data point.

    The errors found by calculating the standard deviation of the nopts
    data points closest to that data point.

    Parameters
    --
    y: data - one column for each replicate
    nopts: number of points used to estimate error bars
    """
    y = np.asarray(y)
    if y.ndim == 1:
        ebar = np.empty(len(y))
        if not nopts:
            nopts = np.round(0.1 * len(y))
        for i in range(len(y)):
            ebar[i] = np.std(np.sort(np.abs(y[i] - y))[:nopts])
        return ebar
    else:
        print("estimateerrorbar: works for 1-d arrays only.")


def natural_keys(text):
    """
    Find a set of keys.

    These keys can be used to sort a list of strings by the numeric
    values in the list entries.

    Eg sorted(list, key= natural_keys)

    Parameters
    ---------
    text: a string (but see above example)
    """
    import re

    def atof(text):
        try:
            retval = float(text)
        except ValueError:
            retval = text
        return retval

    return [
        atof(c)
        for c in re.split(r"[+-]?([0-9]+(?:[.][0-9]*)?|[.][0-9]+)", text)
    ]


def islistempty(ell):
    """
    Check if a list of lists is empty, e.g. [[]].

    Parameters
    --
    ell: list
    """
    if isinstance(ell, list):  # Is a list
        return all(map(islistempty, ell))
    return False  # Not a list


def makelist(c):
    """
    Ensure that a variable is a list.

    Parameters
    --
    c: variable to be made into a list
    """
    import numpy as np

    if isinstance(c, np.ndarray):
        return list(c)
    elif type(c) is not list:
        return [c]
    else:
        return c


def figs2pdf(savename):
    """Save all open figures to a pdf file."""
    from matplotlib.backends.backend_pdf import PdfPages
    import matplotlib.pyplot as plt

    if "." not in savename:
        savename += ".pdf"
    with PdfPages(savename) as pp:
        for i in plt.get_fignums():
            plt.figure(i)
            pp.savefig()



"""Functions for general administration, mostly of data frames."""

import numpy as np
import pandas as pd

import omniplate.omgenutils as gu


def initialiseprogress(self, experiment):
    """Initialise progress dictionary."""
    self.progress["ignoredwells"][experiment] = []
    self.progress["negativevalues"][experiment] = False


def makewellsdf(df_r):
    """Make a dataframe with the contents of the wells."""
    df = df_r[["experiment", "condition", "strain", "well"]].drop_duplicates()
    df = df.reset_index(drop=True)
    return df


def drop(self, experiments, conditions, strains):
    """
    Drop strains from s and sc data frames.

    Variables
    ---------
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    """
    experiments = gu.makelist(experiments)
    conditions = gu.makelist(conditions)
    strains = gu.makelist(strains)
    for attr in ["s", "sc"]:
        df = getattr(self, attr)
        indices_to_drop = [
            df[
                (df.experiment == e) & (df.condition == c) & (df.strain == s)
            ].index
            for e, c, s in zip(experiments, conditions, strains)
        ]
        if indices_to_drop and any(len(idx) > 0 for idx in indices_to_drop):
            getattr(self, attr).drop(
                pd.Index(np.concatenate(indices_to_drop)), inplace=True
            )


def make_s(self, tmin=None, tmax=None, rdf=None):
    """
    Generate s dataframe.

    Calculates means and variances of all data types from raw data.

    Drop "original_experiment" and "experiment_id" because there should
    be one mean_OD for all experiment_ids.
    """
    if rdf is None:
        # restrict time
        if tmin and not tmax:
            rdf = self.r[self.r.time >= tmin]
        elif tmax and not tmin:
            rdf = self.r[self.r.time <= tmax]
        elif tmin and tmax:
            rdf = self.r[(self.r.time >= tmin) & (self.r.time <= tmax)]
        else:
            rdf = self.r
    # find indices to use for s dataframe
    dfindices = self.r.select_dtypes(include=["object"]).columns.tolist()
    for column in ["well", "original_experiment", "experiment_id"]:
        if column in dfindices:
            dfindices.remove(column)
    # classify columns
    groupby_columns = dfindices + ["time"]
    datatypes = self.r.select_dtypes(include="number").columns.tolist()
    good_columns = groupby_columns + ["well"] + datatypes
    # for common variables
    good_columns += [field for field in rdf.columns if "common" in field]
    # find and drop remaining columns
    bad_columns = [col for col in rdf.columns if col not in good_columns]
    if bad_columns:
        rdf = rdf.drop(columns=bad_columns)
    # find means
    df1 = rdf.groupby(groupby_columns).mean(numeric_only=True).reset_index()
    for exp in self.allexperiments:
        for dtype in self.datatypes[exp]:
            df1 = df1.rename(columns={dtype: "mean_" + dtype})
    # find std
    df2 = rdf.groupby(groupby_columns).std(numeric_only=True).reset_index()
    for exp in self.allexperiments:
        for dtype in self.datatypes[exp]:
            df2 = df2.rename(columns={dtype: dtype + "_err"})
    return pd.merge(df1, df2)


def update_s(self):
    """Update means and errors of all datatypes from raw data."""
    # find tmin and tmax in case restrict_time has been called
    tmin = self.s.time.min()
    tmax = self.s.time.max()
    # recalculate s dataframe
    self.s = make_s(self, tmin, tmax)


def add_to_s(self, derivname, outdf):
    """
    Add dataframe of time series to s dataframe.

    Parameters
    ----------
    derivname: str
        Root name for statistic described by dataframe, such as "gr".
    outdf: dataframe
        Data to add.
    """
    gu.merge_df_into(
        self.s, outdf, ["experiment", "condition", "strain", "time"]
    )
    # to prevent fragmentation
    self.s = self.s.copy()


def add_dict_to_sc(self, statsdict):
    """Add one-line dict to sc dataframe."""
    statsdf = pd.DataFrame(statsdict, index=pd.RangeIndex(0, 1, 1))
    gu.merge_df_into(self.sc, statsdf, ["experiment", "condition", "strain"])
    # to prevent fragmentation
    self.sc = self.sc.copy()


def check_kwargs(kwargs):
    """Stop if final s missing from experiments, conditions, or strains."""
    if "condition" in kwargs:
        raise SystemExit("Use conditions not condition as an argument.")
    elif "strain" in kwargs:
        raise SystemExit("Use strains not strain as an argument.")
    elif "experiment" in kwargs:
        raise SystemExit("Use experiments not experiment as an argument.")


@property
def cols_to_underscore(self):
    """Replace spaces in column names of all dataframes with underscores."""
    for df in [self.r, self.s, self.sc]:
        df.columns = df.columns.str.replace(" ", "_")



import numpy as np


def parseSunrise(dfd, rdict):
    """Parse a data frame from a Sunrise plate reader into a dict."""
    rdict["OD"] = []
    # extract times of measurements
    t = (
        np.array([float(str(ts).split("s")[0]) for ts in dfd.to_numpy()[0]])
        / 3600
    )
    t = t[~np.isnan(t)]
    # extract data
    for x in np.arange(1, 13):
        for y in "ABCDEFGH":
            well = y + str(x)
            if well in dfd["Well positions"].values:
                data = dfd[dfd["Well positions"] == well].iloc[:, :-1].values
                data = data.reshape(data.size)
                for i, tv in enumerate(t):
                    rdict["time"].append(tv)
                    rdict["well"].append(well)
                    rdict["OD"].append(data[i])
    return rdict



"""Functions to display the contents of wells."""

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omerrors as errors
import omniplate.omgenutils as gu
import omniplate.sunder as sunder


def contentsofwells(self, wlist):
    """
    Display contents of wells.

    Parameters
    ----------
    wlist: string or list of string
        Specifies the well or wells of interest.

    Examples
    --------
    >>> p.contentsofwells(['A1', 'E4'])
    """
    wlist = gu.makelist(wlist)
    for w in wlist:
        print("\n" + w + "\n--")
        print(
            self.wellsdf.query("well == @w")
            .drop(["well"], axis=1)
            .to_string(index=False)
        )


def showwells(
    self,
    concise=False,
    sortby=True,
    experiments="all",
    conditions="all",
    strains="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditionincludes=False,
    conditionexcludes=False,
    strainincludes=False,
    strainexcludes=False,
):
    """
    Display wells for specified experiments, conditions, and strains.

    Parameters
    ----------
    concise: boolean
        If True, display as experiment: condition: strain.
    sortby: boolean or list of strings, optional
        If True, a default sort will be used.
        If list of column names, sort on these names.
        If False, no sorting will occur.
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in their
        name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.

    Examples
    --------
    >>> p.showwells()
    >>> p.showwells(strains= 'Mal12:GFP', conditions= '1% Mal')
    """
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull=False,
    )
    if not hasattr(self, "wellsdf"):
        self.wellsdf = admin.makewellsdf(self.r)
    df = self.wellsdf.query(
        "experiment == @exps and condition == @cons and strain == @strs"
    ).copy()
    print()
    if concise:
        sdf = df[["experiment", "condition", "strain"]]
        ndf = (
            sdf.groupby(sdf.columns.tolist())
            .size()
            .reset_index(name="replicates")
        )
        df = ndf
    if isinstance(sortby, list) or isinstance(sortby, str):
        df = df.sort_values(by=gu.makelist(sortby))
    elif sortby:
        # default sort
        df[["new1", "new2"]] = df.condition.str.split("%", expand=True)[[0, 1]]
        # sort by nutrient then by its concentration then by strain
        if df.new1.str.isnumeric().any():
            df = df.sort_values(by=["new2", "new1", "strain"])
        else:
            df = df.sort_values(by=["new1", "new2", "strain"])
        df.drop(columns=["new1", "new2"], inplace=True)
    for e in exps:
        print(df.query("experiment == @e").to_string(index=False))
        print()


@clogger.log
def ignorewells(self, ignore=None):
    """
    Ignore the wells specified in any future processing.

    If called several times, the default behaviour is for any previously
    ignored wells not to be re-instated.

    Parameters
    ---------
    ignore: list of strings or dict
        If there is only a single experiment, ignore is a list of labels
        of wells to be ignored.
        If there are multiple experiments, ignore is a dict with experiment
        names as keys and a list of wells as items.

    Example
    -------
    >>> p.ignorewells(['A1', 'C2'])
    >>> p.ignorewells({"exp1" : ["A1"], "exp2": ["C3", "C4"]})
    """
    if ignore is None:
        return
    else:
        if len(self.allexperiments) == 1 and not isinstance(ignore, dict):
            # make exclude a dict
            if isinstance(ignore, str):
                ignore = {self.allexperiments[0]: gu.makelist(ignore)}
            elif isinstance(ignore, list):
                ignore = {self.allexperiments[0]: ignore}
            else:
                raise errors.IgnoreWells(f"{ignore} is in the wrong format.")
        if isinstance(ignore, dict):
            # make each value a list
            ignore = {
                key: [value] if isinstance(value, str) else value
                for key, value in ignore.items()
            }
            if self.allexperiments == self.combined:
                ignore = convert_wells_for_combined(self, ignore)
            # drop wells
            for exp, wells in ignore.items():
                # wells cannot be ignored twice
                wex = list(
                    set(wells) - set(self.progress["ignoredwells"][exp])
                )
                # drop data from ignored wells
                df = self.r
                filt = (df["experiment"] == exp) & df["well"].isin(wex)
                df = df.loc[~filt]
                df = df.reset_index(drop=True)
                self.r = df
                # store ignoredwells
                self.progress["ignoredwells"][exp] += wex
                # remove any duplicates
                self.progress["ignoredwells"][exp] = list(
                    set(self.progress["ignoredwells"][exp])
                )
            # remake s data frame
            admin.update_s(self)
        else:
            raise errors.IgnoreWells(f"{ignore} is in the wrong format.")


def convert_wells_for_combined(self, ignore: dict):
    """Convert ignore into a dict with one key for a combined experiment."""
    if self.allexperiments != self.combined:
        return ignore
    else:
        emap = self.experiment_map
        cignore = {
            self.combined: [
                f"{emap[key]}_{well}" for key in ignore for well in ignore[key]
            ]
        }
        return cignore



"""Functions to plot from the data frames."""

import colorcet
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

import omniplate.admin as admin
import omniplate.clogger as clogger
import omniplate.omerrors as errors
import omniplate.omgenutils as gu
import omniplate.sunder as sunder


@clogger.log
def plot(
    self,
    x="time",
    y="OD",
    hue="strain",
    style="condition",
    size=None,
    kind="line",
    col=None,
    row=None,
    height=5,
    aspect=1,
    xlim=None,
    ylim=None,
    figsize=False,
    returnfacetgrid=False,
    title=None,
    plate=False,
    wells=False,
    nonull=False,
    messages=False,
    sortby=False,
    distinct_colours=False,
    tmin=None,
    tmax=None,
    prettify_dict=None,
    experiments="all",
    conditions="all",
    strains="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditionincludes=False,
    conditionexcludes=False,
    strainincludes=False,
    strainexcludes=False,
    **kwargs,
):
    """
    Plot from the underlying dataframes (chosen automatically).

    Seaborn's relplot is used, which is described at
    https://seaborn.pydata.org/generated/seaborn.relplot.html

    Parameters
    ----------
    x: string
        The variable - column of the dataframe - for the x-axis.
    y: string
        The variable - column of the dataframe - for y-axis.
    hue: string
        The variable whose variation will determine the colours of the
        lines plotted. From Seaborn.
    style: string
        The variable whose variation will determine the style of each line.
        From Seaborn.
    size: string
        The variable whose vairation will determine the size of each
        marker. From Seaborn.
    kind: string
        Either 'line' or 'scatter', which determines the type of plot.
        From Seaborn.
    col: string, optional
        The variable that varies over the columns in a multipanel plot.
        From Seaborn.
    row: string, optional
        The variable that varies over the rows in a multipanel plot.
        From Seaborn.
    height: float, optional
        The height of the individual panels in a multipanel plot.
        From Seaborn.
    aspect: float, optional
        The aspect ratio of the individual panels in a multipanel plot.
        From Seaborn.
    xlim: list of two floats, optional
        The minimal and maximal x-value, such as [0, None]
    ylim: list of two floats, optional
        The minimal and maximal y-value, such as [0, None]
    figsize: tuple, optional
        A tuple of (width, height) for the size of figure.
        Ignored if wells= True or plate= True.
    returnfacetgrid: boolean, optional
        If True, return Seaborn's facetgrid object created by relplot
    title: float, optional
        The title of the plot (overwrites any default titles).
    plate: boolean, optional
        If True, data for each well for a whole plate are plotted in one
        figure.
    wells: boolean, optional
        If True, data for the individual wells is shown.
    nonull: boolean, optional
        If True, 'Null' strains are not plotted.
    sortby: list of strings, optional
        A list of columns to sort the data in the dataframe and passed to
        pandas sort_values.
    messsages: boolean, optional
        If True, print warnings for any data requested but not found.
    distinct_colours: boolean, optional
        If True, try to make neighbouring colours in the plot distinct
        rather than graded.
    tmin: float, optional
        If specifed, restrict the data to times greater than tmin.
    tmax: float, optional
        If specifed, restrict the data to times less than tmax.
    prettify_dict: dict, optional
        To replace the x- and y-axis labels:
            e.g., {"time": "time (h)", "OD": "optical density"}
    experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in
        their name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.
    kwargs: for Seaborn's relplot
        https://seaborn.pydata.org/generated/seaborn.relplot.html

    Returns
    -------
    sfig: Seaborn's facetgrid object generated by relplot if
    returnfacetgrid= True

    Examples
    --------
    >>> p.plot(y= 'OD', plate= True)
    >>> p.plot(y= 'OD', wells= True, strainincludes= 'Gal10:GFP')
    >>> p.plot(y= 'OD')
    >>> p.plot(x= 'OD', y= 'gr')
    >>> p.plot(y= 'cGFPperOD', nonull= True, ymin= 0)
    >>> p.plot(y= 'cGFPperOD', conditionincludes= '2% Mal',
    ...        hue= 'strain')
    >>> p.plot(y= 'cmCherryperOD', conditions= ['0.5% Mal',
    ...        '1% Mal'], hue= 'strain', style= 'condition',
    ...         nonull= True, strainincludes= 'mCherry')
    >>> p.plot(y= 'cGFPperOD', col= 'experiment')
    >>> p.plot(y= 'max gr')
    """
    admin.check_kwargs(kwargs)
    # choose the correct dataframe
    basedf, dfname = plotfinddf(self, x, y, tmin, tmax)
    # get experiments, conditions and strains
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull,
    )
    # check if default prettify_dict
    if prettify_dict is None and hasattr(self, "prettify_dict"):
        prettify_dict = getattr(self, "prettify_dict")
    # choose the right type of plot
    if plate:
        dtype = y if x == "time" else x
        plotplate(self, basedf, exps, dtype)
    elif wells:
        plot_wells(
            x,
            y,
            basedf,
            exps,
            cons,
            strs,
            style=style,
            size=size,
            kind=kind,
            col=col,
            row=row,
            xlim=xlim,
            ylim=ylim,
            title=title,
            figsize=figsize,
            messages=messages,
            distinct_colours=distinct_colours,
            prettify_dict=prettify_dict,
            **kwargs,
        )
    elif dfname == "s" or dfname == "r":
        sfig = plot_rs(
            x,
            y,
            basedf,
            exps,
            cons,
            strs,
            hue=hue,
            style=style,
            size=size,
            kind=kind,
            col=col,
            row=row,
            height=height,
            aspect=aspect,
            xlim=xlim,
            ylim=ylim,
            title=title,
            figsize=figsize,
            sortby=sortby,
            returnfacetgrid=returnfacetgrid,
            distinct_colours=distinct_colours,
            prettify_dict=prettify_dict,
            **kwargs,
        )
        if returnfacetgrid:
            return sfig
    elif dfname == "sc":
        plot_sc(
            x,
            y,
            basedf,
            exps,
            cons,
            strs,
            hue=hue,
            style=style,
            size=size,
            kind=kind,
            col=col,
            row=row,
            height=height,
            aspect=aspect,
            xlim=xlim,
            ylim=ylim,
            figsize=figsize,
            title=title,
            sortby=sortby,
            distinct_colours=distinct_colours,
            prettify_dict=prettify_dict,
            **kwargs,
        )
    else:
        raise errors.PlotError("omplot: No data found")


def plotplate(self, basedf, exps, dtype):
    """
    Plot the data for each well following the layout of a 96-well plate.

    Parameters
    ----------
    self: platereader object
    basedf: DataFrame
        The r dataframe.
    exps: float
        The name of the experiments.
    dtype: float
        The data type to be plotted: 'OD', 'GFP', etc.
    """
    if exps == self.combined:
        exps = list(self.r.original_experiment.unique())
        experiment_column = "original_experiment"
    else:
        experiment_column = "experiment"
    for e in exps:
        plt.figure()
        # first create an empty plate - in case of missing wells
        ax = []
        for rowl in range(8):
            for coll in np.arange(1, 13):
                sindex = coll + 12 * rowl
                axi = plt.subplot(8, 12, sindex)
                ax.append(axi)
                plt.tick_params(labelbottom=False, labelleft=False)
                # label well locations
                for j in range(12):
                    if sindex == j + 1:
                        plt.title(j + 1)
                for j, k in enumerate(np.arange(1, 96, 12)):
                    if sindex == k:
                        plt.ylabel("ABCDEFGH"[j] + " ", rotation=0)
        # fill in the wells that have been measured
        for pl in basedf.query(f"{experiment_column} == @e")["well"].unique():
            if experiment_column == "experiment":
                well_loc = pl
            else:
                well_loc = pl.split("_")[1]
            rowl = "ABCDEFGH".index(well_loc[0])
            coll = int(well_loc[1:])
            sindex = coll + 12 * rowl
            wd = basedf.query(f"{experiment_column} == @e and well == @pl")
            ax[sindex - 1].plot(
                wd["time"].to_numpy(), wd[dtype].to_numpy(), "-"
            )
        plt.suptitle(e + ": " + dtype)
        plt.show(block=False)


def plot_wells_facet_grid(
    basedf,
    x,
    y,
    row,
    col,
    e,
    cons,
    strs,
    prettify_dict,
    title,
    xlim,
    ylim,
    figsize,
):
    """Use facetgrid to show multiple plots simultaneously."""
    df = basedf.query(
        "experiment == @e and condition == @cons and strain == @strs"
    )
    if row is not None and col is not None:
        groups = [row, col]
    elif row is not None:
        groups = [row]
    elif col is not None:
        groups = [col]
    sfig = sns.FacetGrid(df, row=row, col=col)
    for var, facet_df in df.groupby(groups):
        if row is not None and col is not None:
            row_var, col_var = var
            ax = sfig.axes[
                sfig.row_names.index(row_var),
                sfig.col_names.index(col_var),
            ]
        elif row is not None:
            ax = sfig.axes[sfig.row_names.index(var[0])][0]
        elif col is not None:
            ax = sfig.axes[sfig.col_names.index(var[0])][0]
        sns.lineplot(x=x, y=y, hue="well", data=facet_df, ax=ax)
        ax.set(xlabel="", ylabel="")
    sfig.set_titles()
    if title is None:
        title = e
    set_labels_limits(sfig, x, y, title, xlim, ylim, prettify_dict)
    if figsize and len(figsize) == 2:
        sfig.fig.set_figwidth(figsize[0])
        sfig.fig.set_figheight(figsize[1])
    plt.tight_layout()
    plt.show(block=False)


def plot_wells_multiple(
    basedf,
    x,
    y,
    kind,
    style,
    size,
    e,
    cons,
    strs,
    prettify_dict,
    xlim,
    ylim,
    distinct_colours,
    messages,
    kwargs,
):
    """Create plots for each strain and condition."""
    for c in cons:
        for s in strs:
            df = basedf.query(
                "experiment == @e and condition == @c and strain == @s"
            )
            if df.empty:
                if messages:
                    print(e + ":", "No data found for", s, "in", c)
            else:
                if distinct_colours:
                    palette = sns.color_palette(
                        colorcet.glasbey, df.well.unique().size
                    )
                else:
                    palette = None
                sfig = sns.relplot(
                    x=x,
                    y=y,
                    data=df,
                    hue="well",
                    kind=kind,
                    style=style,
                    size=size,
                    palette=palette,
                    **kwargs,
                )
                title = e + ": " + s + " in " + c
                set_labels_limits(sfig, x, y, title, xlim, ylim, prettify_dict)
                plt.show(block=False)


def plot_wells(
    x,
    y,
    basedf,
    exps,
    cons,
    strs,
    style="condition",
    size=None,
    kind="line",
    col=None,
    row=None,
    xlim=None,
    ylim=None,
    title=None,
    figsize=None,
    messages=False,
    distinct_colours=False,
    prettify_dict=None,
    **kwargs,
):
    """
    Plot data from the individual wells.

    Data for each experiment, condition, and strain are plotted in
    a separate figure unless row and col are specified.
    """
    for e in exps:
        if row or col:
            plot_wells_facet_grid(
                baseddf=basedf,
                x=x,
                y=y,
                row=row,
                col=col,
                e=e,
                cons=cons,
                strs=strs,
                prettify_dict=prettify_dict,
                title=title,
                xlim=xlim,
                ylim=ylim,
                figsize=figsize,
            )
        else:
            plot_wells_multiple(
                basedf=basedf,
                x=x,
                y=y,
                kind=kind,
                style=style,
                size=size,
                e=e,
                cons=cons,
                strs=strs,
                prettify_dict=prettify_dict,
                xlim=xlim,
                ylim=ylim,
                distinct_colours=distinct_colours,
                messages=messages,
                kwargs=kwargs,
            )


def print_plot_warnings(hue, style, size, col, x, df):
    """Print warnings if user's choices seem off."""
    if hue == style:
        print(
            f'Warning: "hue" and "style" have both been set to {hue}"'
            '" and there may be unintended averaging.'
        )
    if (
        x != "commontime"
        and len(df["experiment"].unique()) > 1
        and hue != "experiment"
        and size != "experiment"
        and style != "experiment"
        and col != "experiment"
    ):
        print(
            "Warning: there are multiple experiments, but neither "
            '"hue", "style", nor "size" is set to "experiment" and there'
            " may be averaging over experiments."
        )


def plot_rs(
    x,
    y,
    basedf,
    exps,
    cons,
    strs,
    hue="strain",
    style="condition",
    size=None,
    kind="line",
    col=None,
    row=None,
    height=5,
    aspect=1,
    xlim=None,
    ylim=None,
    title=None,
    figsize=None,
    sortby=False,
    returnfacetgrid=False,
    distinct_colours=False,
    prettify_dict=None,
    **kwargs,
):
    """Plot time-series data from the .r or .s dataframes."""
    # plot time series
    df = basedf.query(
        "experiment == @exps and condition == @cons and strain == @strs"
    )
    if df.empty or df.isnull().all().all():
        # no data or data all NaN
        print("omplot: No data found.")
    else:
        if sortby:
            df = df.sort_values(by=gu.makelist(sortby))
        # warn if poor choice of seaborn's parameters
        print_plot_warnings(hue, style, size, col, x, df)
        if kind == "line" and "units" not in kwargs:
            # augment df to allow seaborn to estimate errors
            df = augmentdf(df, y)
            kwargs["errorbar"] = "sd"
        else:
            errors.PlotError(f"omplot: kind={kind} is not recognised.")
        # plot
        if distinct_colours:
            palette = sns.color_palette(
                colorcet.glasbey, df[hue].unique().size
            )
            kwargs["palette"] = palette
        # remove conditions and strains that are all nan
        cdf = (
            df.groupby(["experiment", "condition", "strain"])[y]
            .apply(np.nanmean)
            .reset_index()
        )
        keep_conditions = list(cdf.dropna().condition.values)
        keep_strains = list(cdf.dropna().strain.values)
        # use seaborn to plot
        tdf = df[
            df.condition.isin(keep_conditions) & df.strain.isin(keep_strains)
        ]
        if not tdf.empty:
            sfig = sns.relplot(
                x=x,
                y=y,
                data=tdf,
                hue=hue,
                kind=kind,
                style=style,
                size=size,
                col=col,
                row=row,
                aspect=aspect,
                height=height,
                **kwargs,
            )
            set_labels_limits(sfig, x, y, title, xlim, ylim, prettify_dict)
            if figsize and len(figsize) == 2:
                sfig.fig.set_figwidth(figsize[0])
                sfig.fig.set_figheight(figsize[1])
            plt.show(block=False)
            if returnfacetgrid:
                return sfig
            else:
                return None
        else:
            print("omplot: No data found.")


def plot_sc(
    x,
    y,
    basedf,
    exps,
    cons,
    strs,
    hue="strain",
    style="condition",
    size=None,
    kind="scatter",
    col=None,
    row=None,
    height=5,
    aspect=1,
    xlim=None,
    ylim=None,
    figsize=None,
    title=None,
    sortby=False,
    distinct_colours=False,
    prettify_dict=None,
    **kwargs,
):
    """Plot summary statistics from the .sc dataframe."""
    # plot summary stats
    df = basedf.query(
        "experiment == @exps and condition == @cons and strain == @strs"
    )
    xcols = df.columns[df.columns.str.startswith(x)]
    ycols = df.columns[df.columns.str.startswith(y)]
    cols_to_keep = (
        ["experiment", "condition", "strain"] + list(xcols) + list(ycols)
    )
    sortby = gu.makelist(sortby) if sortby else []
    for field in [hue, style, size, *sortby, col, row]:
        if isinstance(field, str):
            cols_to_keep += [field]
    df = df[np.unique(cols_to_keep)].dropna()
    if df.empty or df.isnull().all().all():
        # no data or data all NaN:
        print("omplot: No data found.")
    else:
        if sortby:
            df = df.sort_values(by=gu.makelist(sortby))
        if distinct_colours:
            kwargs["palette"] = sns.color_palette(
                colorcet.glasbey, df[hue].unique().size
            )
        sfig = sns.relplot(
            x=x,
            y=y,
            data=df,
            hue=hue,
            style=style,
            size=size,
            col=col,
            row=row,
            aspect=aspect,
            height=height,
            **kwargs,
        )
        set_labels_limits(sfig, x, y, title, xlim, ylim, prettify_dict)
        if row is None and col is None:
            # add error bars
            # find coordinates of points in relplot
            xc, yc = [], []
            for point_pair in sfig.ax.collections:
                for xp, yp in point_pair.get_offsets():
                    xc.append(xp)
                    yc.append(yp)
            # add error bars
            xerr = df[x + "_err"] if x + "_err" in df.columns else None
            yerr = df[y + "_err"] if y + "_err" in df.columns else None
            sfig.ax.errorbar(
                xc,
                yc,
                xerr=xerr,
                yerr=yerr,
                fmt=" ",
                ecolor="dimgray",
                alpha=0.5,
            )
        plt.show(block=False)


def plotfinddf(self, x, y, tmin, tmax):
    """
    Find the correct dataframe for plotting y versus x.

    Parameters
    ----------
    self: a platereader instance
    x: string
        Name of x-variable.
    y: string
        Name of y-variable.
    tmin: float
        If specifed, restrict the data to times greater than tmin.
    tmax: float
        If specifed, restrict the data to times less than tmax.

    Returns
    -------
    basedf: dataframe
        The dataframe that contains the x and y variables.
    dfname: string
        The name of the dataframe.
    """
    # choose the correct dataframe
    if hasattr(self, "r") and x in self.r.columns and y in self.r.columns:
        # raw data (with wells)
        basedf = self.r
        dfname = "r"
    elif x in self.s.columns and y in self.s.columns:
        # processed data (no wells)
        basedf = self.s
        dfname = "s"
    elif x in self.sc.columns and y in self.sc.columns:
        # summary stats
        basedf = self.sc
        dfname = "sc"
    else:
        raise errors.PlotError(
            f"The variables x= {x}"
            + f" and y= {y}"
            + " cannot be plotted against each other because they are not in "
            + " the same dataframe"
        )
    if (tmin or tmax) and "time" in basedf.columns:
        if tmin is not None and tmax is None:
            basedf = basedf[basedf.time >= tmin]
        elif tmin is None and tmax is not None:
            basedf = basedf[basedf.time <= tmax]
        elif tmin is not None and tmax is not None:
            basedf = basedf[(basedf.time >= tmin) & (basedf.time <= tmax)]
    return basedf, dfname


def augmentdf(df, datatype):
    """
    Augment dataframe to allow Seaborn to generate errors.

    Use 'err' (if present in the dataframe) to allow Seaborn to generate
    errors in relplot, otherwise returns original dataframe.

    Note we call seaborn with errorbar = "sd" and so use sqrt(3/2) * error
    because seaborn calculates the standard deviation from the augmented data
    (the mean, the mean + std, and the mean - std) and so gets
    std/sqrt(3/2) otherwise because there are three data points.
    """
    if datatype + "_err" in df:
        derr = datatype + "_err"
    elif "mean" in datatype and datatype.split("mean_")[1] + "_err" in df:
        derr = datatype.split("mean_")[1] + "_err"
    else:
        derr = None
        # return if df is df_r
        return df
    if derr:
        df.insert(0, "augtype", "mean")
        # add errors
        mn = df[datatype].to_numpy()
        err = df[derr].to_numpy()
        # add std
        dfp = df.copy()
        dfp[datatype] = mn + np.sqrt(3 / 2) * err
        dfp["augtype"] = "+err"
        # minus std
        dfm = df.copy()
        dfm[datatype] = mn - np.sqrt(3 / 2) * err
        dfm["augtype"] = "-err"
        # concat
        df = pd.concat([df, dfp, dfm], ignore_index=True)
    return df


def savefigs(self, fname=None, onefile=True):
    """
    Save all current figures to PDF.

    Either all figures save to one file or each to a separate one.

    Parameters
    ----------
    fname: string, optional
        Name of file. If unspecified, the name of the experiment is used.
    onefile: boolean, optional
        If False, each figures is save to its own PDF file.

    Example
    -------
    >>> p.savefigs()
    >>> p.savefigs('figures.pdf')
    """
    if fname:
        if ".pdf" not in fname:
            fname += ".pdf"
        fname = str(self.wdirpath / fname)
    else:
        fname = str(self.wdirpath / ("".join(self.allexperiments) + ".pdf"))
    if onefile:
        gu.figs2pdf(fname)
    else:
        for i in plt.get_fignums():
            plt.figure(i)
            savename = str(plt.getp(plt.gcf(), "axes")[0].title).split("'")[1]
            savename = savename.replace(" ", "_")
            if savename == "":
                savename = "Whole_plate_Figure_" + str(i)
            print("Saving", savename)
            plt.savefig(str(self.wdirpath / (savename + ".pdf")))


@property
def close(self):
    """
    Close all figures.

    Example
    -------
    >>> p.close
    """
    plt.close("all")


def set_labels_limits(sfig, x, y, title, xlim, ylim, prettify_dict):
    """Set axes labels and limits and set title."""
    if prettify_dict is not None:
        sfig.set_axis_labels(prettify_dict.get(x, x), prettify_dict.get(y, y))
    else:
        sfig.set_axis_labels(x, y)
    if xlim is not None:
        plt.xlim(xlim)
    if ylim is not None:
        plt.ylim(ylim)
    if title:
        sfig.fig.suptitle(title)


def inspect(
    self,
    xlabel="time",
    ylabel="bcGFPperOD",
    refstrain="WT",
    fl="GFP",
    max_xy=None,
    experiments="all",
    conditions="all",
    strains="all",
    experimentincludes=False,
    experimentexcludes=False,
    conditionincludes=False,
    conditionexcludes=False,
    strainincludes=False,
    strainexcludes=False,
):
    """
    Plot growth and fluorescence data in one plot per strain per condition.

    Parameters
    ----------
    xlabel: string
        Variable to plot on the x-axis.
    ylabel: string
        Variable to plot on the y-axis.
    refstrain: string
        The reference strain used to correct autofluorescence.
    fl: string
        The name of the fluorescence channel.
    max_xy: (float, float)
        A maximum to show on the plot of xlabel versus ylabel.
     experiments: string or list of strings
        The experiments to include.
    conditions: string or list of strings
        The conditions to include.
    strains: string or list of strings
        The strains to include.
    experimentincludes: string, optional
        Selects only experiments that include the specified string in
        their name.
    experimentexcludes: string, optional
        Ignores experiments that include the specified string in their
        name.
    conditionincludes: string, optional
        Selects only conditions that include the specified string in their
        name.
    conditionexcludes: string, optional
        Ignores conditions that include the specified string in their name.
    strainincludes: string, optional
        Selects only strains that include the specified string in their
        name.
    strainexcludes: string, optional
        Ignores strains that include the specified string in their name.
    """
    exps, cons, strs = sunder.getall(
        self,
        experiments,
        experimentincludes,
        experimentexcludes,
        conditions,
        conditionincludes,
        conditionexcludes,
        strains,
        strainincludes,
        strainexcludes,
        nonull=True,
    )
    flperod = f"flperOD_raw_{fl}"
    for exp in exps:
        for con in cons:
            for strain in strs:
                sdf, x, y, midlog_times, n_plots, ref_df = set_up_inspect(
                    self, exp, con, strain, xlabel, ylabel, flperod, refstrain
                )
                if x.size == 0 or y.size == 0:
                    print(
                        f"{strain} in {con}: No data found for"
                        f" {xlabel} and {ylabel}."
                    )
                    continue
                # begin plot
                curr_plot = 1
                plt.figure(figsize=(5, 7))
                # plot OD
                ax1 = plt.subplot(n_plots, 1, curr_plot)
                plt.plot(x, sdf["mean_OD"].values, "m")
                plt.plot(
                    ref_df["time"].values,
                    ref_df["mean_OD"].values,
                    "k--",
                    alpha=0.5,
                    label=refstrain,
                )
                if midlog_times is not None and xlabel == "time":
                    add_midlog_lines(
                        x,
                        midlog_times,
                        sdf["mean_OD"].values,
                    )
                plt.ylabel("OD")
                plt.legend()
                plt.grid()
                curr_plot += 1
                # plot growth rate
                if "gr" in sdf.columns:
                    # growth rate
                    plt.subplot(n_plots, 1, curr_plot, sharex=ax1)
                    plt.plot(x, sdf["gr"].values, "m")
                    if midlog_times and xlabel == "time":
                        add_midlog_lines(x, midlog_times, sdf["gr"].values)
                    plt.ylabel("gr")
                    plt.grid()
                    curr_plot += 1
                # ylabel plot
                plt.subplot(n_plots, 1, curr_plot, sharex=ax1)
                # add inferred GFP per OD if present
                if ylabel == f"bc{fl}perOD" and flperod in sdf.columns:
                    plt.plot(x, sdf[flperod].values, "g.")
                # plot ylabel vs. xlabel
                plt.plot(x, y, "r")
                if max_xy is not None and max_xy[0] is not None:
                    plt.plot(max_xy[0], max_xy[1], "ro")
                if midlog_times and xlabel == "time":
                    add_midlog_lines(x, midlog_times, y)
                plt.ylabel(ylabel)
                plt.grid()
                curr_plot += 1
                # plot bulk GFP
                plt.subplot(n_plots, 1, curr_plot, sharex=ax1)
                plot_bulk_GFP(
                    self,
                    xlabel,
                    fl,
                    strain,
                    con,
                    exp,
                    refstrain,
                    midlog_times,
                )
                # tidy up
                plt.grid()
                plt.xlabel(xlabel)
                plt.suptitle(f"{exp}: {strain} in {con}")
                plt.tight_layout()
                plt.show(block=False)


def set_up_inspect(self, exp, con, strain, xlabel, ylabel, flperod, refstrain):
    """Get data for inspect."""
    sdf = self.s[
        (self.s.experiment == exp)
        & (self.s.strain == strain)
        & (self.s.condition == con)
    ]
    x = sdf[xlabel].values
    y = sdf[ylabel].values
    if "min_midlog_time" in self.sc.columns:
        midlog_times = get_midlog_times(self, strain, con, sdf)
    else:
        midlog_times = None
    # find number of plots
    n_plots = 2
    if flperod in sdf.columns:
        n_plots += 1
    if "gr" in sdf.columns:
        n_plots += 1
    # reference strain
    ref_df = self.s[
        (self.s.experiment == sdf.experiment.unique()[0])
        & (self.s.strain == refstrain)
        & (self.s.condition == sdf.condition.unique()[0])
    ]
    return sdf, x, y, midlog_times, n_plots, ref_df


def plot_bulk_GFP(
    self, xlabel, fl, strain, condition, experiment, refstrain, midlog_times
):
    """Plot bulk GFP for strain of interest and refstrain."""
    rdf = self.r[
        (self.r.experiment == experiment)
        & self.r.strain.isin(["Null", refstrain, strain])
        & (self.r.condition == condition)
    ]
    # GFP-tagged strain
    plt.plot(
        rdf[rdf.strain == strain][xlabel].values,
        rdf[rdf.strain == strain][f"{fl}"].values,
        "g.",
    )
    plt.plot(
        rdf[rdf.strain == refstrain][xlabel].values,
        rdf[rdf.strain == refstrain][f"{fl}"].values,
        "kx",
        alpha=0.5,
        label=refstrain,
    )
    # background - Null wells
    plt.plot(
        rdf[rdf.strain == "Null"][xlabel].values,
        rdf[rdf.strain == "Null"][f"{fl}"].values,
        "m*",
        label="Null",
    )
    if midlog_times and xlabel == "time":
        add_midlog_lines(
            rdf[rdf.strain == strain][xlabel].values,
            midlog_times,
            rdf[rdf.strain == strain][f"{fl}"].values,
        )
    plt.ylabel(fl)
    plt.legend()


def get_midlog_times(p, strain, condition, sdf):
    """Find start and end times of midlog growth."""
    sscdf = p.sc[(p.sc.strain == strain) & (p.sc.condition == condition)]
    t0 = sscdf.min_midlog_time.values[0]
    t1 = sscdf.max_midlog_time.values[0]
    i_t0 = np.argmin((sdf.time.values - t0) ** 2)
    i_t1 = np.argmin((sdf.time.values - t1) ** 2)
    if i_t0 == i_t1:
        print(
            f"inspect: {strain} in {condition} has a "
            "midlog region of size zero."
        )
    return i_t0, i_t1


def add_midlog_lines(x, midlog_times, y):
    """Bound midlog region with dashed lines."""
    i_t0 = midlog_times[0]
    i_t1 = midlog_times[1]
    plt.plot([x[i_t0], x[i_t0]], [0, np.nanmax(y)], "k:")
    plt.plot([x[i_t1], x[i_t1]], [0, np.nanmax(y)], "k:")



import gaussianprocessderivatives as gp
import matplotlib.pylab as plt
import numpy as np
from scipy import integrate
from scipy.interpolate import interp1d

import omniplate.clogger as clogger
import omniplate.omgenutils as gu
import omniplate.sunder as sunder
from omniplate.corrections import gethypers, instantiateGP
from omniplate.omstats import statserr


@clogger.log
def averageoverexpts(
    self,
    condition,
    strain,
    tvr="mean_OD",
    bd=False,
    addnoise=True,
    plot=False,
):
    """
    Average a time-dependent variable over all experiments.

    Uses a Matern Gaussian process.

    An alternative and best first choice is addcommonvar.

    Parameters
    ----------
    condition: string
        The condition of interest.
    strain: string
        The strain of interest.
    tvr: float
        The time-dependent variable to be averaged.
        For example, 'cGFPperOD' or 'mean_OD'.
    bd: dictionary, optional
        The limits on the hyperparameters for the Matern Gaussian process.
        For example, {0: (-5,5), 1: (-4,4), 2: (-5,2)}
        where the first element controls amplitude, setting the bounds to
        1e-5 and 1e5, the second controls flexibility, and the third
        determines the magnitude of the measurement error.
    addnoise: boolean
        If True, add the fitted magnitude of the measurement noise to the
        predicted standard deviation for better comparison with the spread
        of the data.

    Returns
    -------
    res: dictionary
        {'t' : time, tvr : time-dependent data, 'mn' : mean,
        'sd' : standard deviation}
        where 'mn' is the average found and 'sd' is its standard deviation.
        'tvr' is the data used to find the average.

    Examples
    --------
    >>> p.averageoverexpts('1% Gal', 'GAL2', bd= {1: [-1,-1])})
    """
    # boundaries on hyperparameters
    if "OD" in tvr:
        bds = {0: (-4, 4), 1: (-1, 4), 2: (-6, 2)}
    else:
        bds = {0: (2, 12), 1: (-1, 4), 2: (4, 10)}
    if bd:
        bds = gu.mergedicts(original=bds, update=bd)
    # extract data
    df = self.s[["experiment", "condition", "strain", "time", tvr]]
    ndf = df.query("condition == @condition and strain == @strain")
    # use GP to average over experiments
    x = ndf["time"].to_numpy()
    y = ndf[tvr].to_numpy()
    ys = y[np.argsort(x)]
    xs = np.sort(x)
    g = gp.maternGP(bds, xs, ys)
    print(f"Averaging over {tvr} experiments for {strain} in {condition}.")
    g.findhyperparameters(noruns=2, noinits=1000)
    g.results()
    g.predict(xs, addnoise=addnoise)
    if plot:
        plt.figure()
        g.sketch(".")
        plt.title("averaging " + strain + " in " + condition)
        plt.xlabel("time")
        plt.ylabel(tvr)
        plt.show(block=False)
    # return results as a dictionary
    res = {"t": xs, tvr: ys, "mn": g.f, "sd": np.sqrt(g.fvar)}
    return res


def getfitnesspenalty(
    self,
    ref,
    com,
    yvar="gr",
    figs=True,
    nosamples=100,
    abs=False,
    norm=False,
):
    """
    Estimate the difference in fitness between two strains.

    Calculate the area between typically two growth rate versus OD
    curves, normalised by the length along the OD-axis where they overlap.

    Parameters
    -----------
    self: omniplate.platereader instance
        An instance of platereader with data loaded.
    ref: list of strings
        For only a single experiment, a list of two strings. The first string
        specifies the condition and the second specifies the strain to be used
        for the reference to which fitness is to be calculated.
        With multiple experiments, a list of three strings. The first string
        specifies the experiment, the second specifies the condition, and the
        third specifies the strain.
    com: list of strings
        For only a single experiment, a list of two strings. The first string
        specifies the condition and the second specifies the strain to be
        compared with the reference.
        With multiple experiments, a list of three strings. The first string
        specifies the experiment, the second specifies the condition, and the
        third specifies the strain.
    yvar: string, optional
        The variable to be compared.
    figs: boolean, optional
        If True, a plot of the area between the two growth rate versus OD
        curves is shown.
    nosamples: integer
        The number bootstraps used to estimate the error.
    abs: boolean
        If True, integrate the absolute difference between the two curves.
    norm: boolean
        If True, returns the mean and variance of the area under the reference
        strain for normalisation.

    Returns
    -------
    fp: float
        The area between the two curves.
    err: float
        An estimate of the error in the calculated error, found by
        bootstrapping.
    reffp: float, optional
        The area beneath the reference strain.
    referr: float, optional
        An estimate of the error in the calculated area for the reference
        strain.

    Example
    -------
    >>> import omniplate as om
    >>> from om_code.getfitnesspenalty import getfp
    >>> p = om.platereader("ExampleData.xlsx", "ExampleDataContents.xlsx",
            datadir="data")
    >>> p.getstats(strains="WT", conditions = ["2% Mal", "WT"],
            ["2% Raf", "WT"])
    >>> fp, err = getfp(p, ["2% Mal", "WT"], ["2% Raf", "WT"])
    """
    if len(ref) == 2 & len(com) == 2:
        # add experiment
        ref.insert(0, self.allexperiments[0])
        com.insert(0, self.allexperiments[0])
    # get and sample from GPs
    if nosamples and yvar == "gr":
        # instantiate GPS to estimate errors
        for ecs in [ref, com]:
            e, c, s = ecs
            hypers, cvfn = gethypers(self, e, c, s, yvar)
            if hypers is None or cvfn is None:
                raise SystemExit(
                    "\nYou must first run getstats or set nosamples= 0."
                )
            t, od = sunder.extractwells(self.r, self.s, e, c, s, "OD")
            if ecs == ref:
                g_ref = instantiateGP(hypers, cvfn, t, od)
            else:
                g_com = instantiateGP(hypers, cvfn, t, od)
        # sample from GPs
        f0s, g0s = g_ref.sample(nosamples, derivs=1)
        f1s, g1s = g_com.sample(nosamples, derivs=1)
        xsref, ysref = np.exp(f0s), g0s
        xscom, yscom = np.exp(f1s), g1s

    else:
        # no estimates of errors
        if nosamples:
            print(
                "Cannot estimate errors - require y= 'gr' and a recently "
                "run getstats"
            )
        xsref = self.s.query(
            "experiment == @ref[0] and condition == @ref[1] and "
            "strain == @ref[2]"
        )["OD mean"][:, None]
        ysref = self.s.query(
            "experiment == @ref[0] and condition == @ref[1] and "
            "strain == @ref[2]"
        )[yvar].to_numpy()[:, None]
        xscom = self.s.query(
            "experiment == @com[0] and condition == @com[1] and "
            "strain == @com[2]"
        )["OD mean"].to_numpy()[:, None]
        yscom = self.s.query(
            "experiment == @com[0] and condition == @com[1] and "
            "strain == @com[2]"
        )[yvar].to_numpy()[:, None]
        if xsref.size == 0 or ysref.size == 0:
            print(f"{ref[0]}: Data missing for {ref[2]} in {ref[1]}")
            return np.nan, np.nan
        elif xscom.size == 0 or yscom.size == 0:
            print(f"{com[0]}: Data missing for {com[2]} in {com[1]}")
            return np.nan, np.nan
    fps = np.zeros(xsref.shape[1])
    nrm = np.zeros(xsref.shape[1])
    samples = zip(
        np.transpose(xsref),
        np.transpose(ysref),
        np.transpose(xscom),
        np.transpose(yscom),
    )
    # process samples
    for j, (xref, yref, xcom, ycom) in enumerate(samples):
        # remove any double values in OD because of OD plateauing
        uxref, uiref = np.unique(xref, return_inverse=True)
        uyref = np.array(
            [
                np.median(yref[np.nonzero(uiref == i)[0]])
                for i in range(len(uxref))
            ]
        )
        uxcom, uicom = np.unique(xcom, return_inverse=True)
        uycom = np.array(
            [
                np.median(ycom[np.nonzero(uicom == i)[0]])
                for i in range(len(uxcom))
            ]
        )
        # interpolate data
        iref = interp1d(uxref, uyref, fill_value="extrapolate", kind="slinear")
        icom = interp1d(uxcom, uycom, fill_value="extrapolate", kind="slinear")
        # find common range of x
        uxi = np.max([uxref[0], uxcom[0]])
        uxf = np.min([uxref[-1], uxcom[-1]])
        # perform integration to find normalised area between curves
        if abs:

            def igrand(x):
                return np.abs(iref(x) - icom(x))

        else:

            def igrand(x):
                return iref(x) - icom(x)

        fps[j] = integrate.quad(igrand, uxi, uxf, limit=100, full_output=1)[
            0
        ] / (uxf - uxi)
        if norm:
            # calculate area under curve of reference strain as a normalisation
            def igrand(x):
                return iref(x)

            nrm[j] = integrate.quad(
                igrand, uxi, uxf, limit=100, full_output=1
            )[0] / (uxf - uxi)
        # an example figure
        if figs and j == 0:
            plt.figure()
            plt.plot(uxref, uyref, "k-", uxcom, uycom, "b-")
            x = np.linspace(uxi, uxf, np.max([len(uxref), len(uxcom)]))
            plt.fill_between(x, iref(x), icom(x), facecolor="red", alpha=0.5)
            plt.xlabel("OD")
            plt.ylabel(yvar)
            plt.legend(
                [
                    f"{ref[0]}: {ref[2]} in {ref[1]}",
                    f"{com[0]}: {com[2]} in {com[1]}",
                ],
                loc="upper left",
                bbox_to_anchor=(-0.05, 1.15),
            )
            plt.show(block=False)
    if norm:
        return (
            np.median(fps),
            statserr(fps),
            np.median(nrm),
            statserr(nrm),
        )
    else:
        return np.median(fps), statserr(fps)
